<?php if (!defined('ROOTPATH')) exit('No direct script access allowed'); ?>
<?php

/**
 * Backlog Defect Plugin for TestRail
 *
 * Copyright TechMatrix Corporation All rights reserved.
 *
 */

class Backlog_defect_plugin extends Defect_plugin
{
	private $_api;
	private $_address;
	private $_api_key;

	private static $_meta = array(
		'author' => 'TechMatrix',
		'version' => '1.1.0 (2023-07-11)',
		'description' => 'Backlog defect plugin for TestRail',
		'can_push' => true,
		'can_lookup' => true,
		'default_config' =>
		'; Backlog 欠陥プラグイン
; 
; 欠陥表示URL：https://<yoursubdomain>.backlog.com/view/%id%
; 欠陥追加URL：https://<yoursubdomain>.backlog.com/add/<yourprojectname>
;
; 欠陥プラグイン詳細設定：
;  * api_keyは Backlog上の[個人設定] > [API] で発行してください。
;  * Backlogの課題の作成者をTestRailユーザーごとに分ける場合は、
;    TestRailの [管理] > [統合] からユーザー変数を追加してください。
;    以下のapi_keyを api_key=%作成したユーザー変数名% として指定し
;    TestRailの [個人設定] > [設定] にてユーザー変数に
;    APIキーを設定してください。
;
[connection]
address=https://<yoursubdomain>.backlog.com/
api_key=XXXXXXXXXXXXXXXXXXXXXXXXXX
'
	);

	public function get_meta()
	{
		return self::$_meta;
	}

	// *********************************************************
	// CONFIGURATION
	// *********************************************************

	public function validate_config($config)
	{
		$ini = ini::parse($config);

		if (!isset($ini['connection']))
		{
			throw new ValidationException('Missing [connection] group');
		}

		$keys = array('address', 'api_key');

		// Check required values for existance
		foreach ($keys as $key)
		{
			if (!isset($ini['connection'][$key]) ||
				!$ini['connection'][$key])
			{
				throw new ValidationException(
					"Missing configuration for key '$key'"
				);
			}
		}

		$address = $ini['connection']['address'];

		// Check whether the address is a valid url (syntax only)
		if (!check::url($address))
		{
			throw new ValidationException('Address is not a valid url');
		}
	}

	public function configure($config)
	{
		$ini = ini::parse($config);
		$this->_address = str::slash($ini['connection']['address']);
		$this->_api_key = $ini['connection']['api_key'];
	}

	// *********************************************************
	// API / CONNECTION
	// *********************************************************

	private function _get_api()
	{
		if ($this->_api)
		{
			return $this->_api;
		}

		$this->_api = new Backlog_api(
			$this->_address,
			$this->_api_key);

		return $this->_api;
	}

	// *********************************************************
	// PUSH
	// *********************************************************

	public function prepare_push($context)
	{
		// Return a form with the following fields/properties
		return array(
			'fields' => array(
				'summary' => array(
					'type' => 'string',
					'label' => '件名',
					'required' => true,
					'size' => 'full'
				),
				'project' => array(
					'type' => 'dropdown',
					'label' => 'プロジェクト',
					'required' => true,
					'remember' => true,
					'cascading' => true,
					'size' => 'compact'
				),
				'issue_type' => array(
					'type' => 'dropdown',
					'label' => '種別',
					'required' => true,
					'remember' => true,
					'size' => 'compact',
					'depends_on' => 'project',
				),
				'priority' => array(
					'type' => 'dropdown',
					'label' => '優先度',
					'remember' => true,
					'required' => true,
					'size' => 'compact'
				),
				'category' => array(
					'type' => 'multiselect',
					'label' => 'カテゴリー',
					'remember' => true,
					'depends_on' => 'project',
					'size' => 'compact'
				),
				'description' => array(
					'type' => 'text',
					'label' => '課題の詳細',
					'rows' => 10
				),
				'attachments' => array(
					'type' => 'dropbox',
					'label' => 'attachments',
					'required' => false,
					'size' => 'none'
				),
			)
		);
	}

	private function _get_summary_default($context)
	{
		$test = current($context['tests']);
		$subject = 'Failed test: ' . $test->case->title;

		if ($context['test_count'] > 1)
		{
			$subject .= ' (+others)';
		}

		return $subject;
	}

	private function _get_description_default($context)
	{
		return $context['test_change']->description;
	}

	private function _to_id_name_lookup($items)
	{
		$result = array();
		if ($items != null && is_array($items)) {
			foreach ($items as $item) {
				$result[$item->id] = $item->name;
			}
		}
		return $result;
	}

	private function _get_projects($api)
	{
		return $this->_to_id_name_lookup($api->get_projects());
	}

	private function _get_issue_types($api, $project_id)
	{
		return $this->_to_id_name_lookup($api->get_issue_types($project_id));
	}

	private function _get_categories($api, $project_id)
	{
		return $this->_to_id_name_lookup($api->get_categories($project_id));
	}

	private function _get_priority($api)
	{
		return $this->_to_id_name_lookup($api->get_priority());
	}

	private function _get_project_from_id($api, $project_id)
	{
		return $api->get_project_from_id($project_id);
	}

	public function prepare_field($context, $input, $field)
	{
		$data = array();

		// Process those fields that do not need a connection to the
		// backlog installation.
		if ($field == 'summary' || $field == 'description')
		{
			switch ($field)
			{
				case 'summary':
					$data['default'] = $this->_get_summary_default(
						$context);
					break;

				case 'description':
					$data['default'] = $this->_get_description_default(
						$context);
					break;
			}

			return $data;
		}

		// Take into account the preferences of the user, but only
		// for the initial form rendering (not for dynamic loads).
		if ($context['event'] == 'prepare')
		{
			$prefs = arr::get($context, 'preferences');
		}
		else
		{
			$prefs = null;
		}

		// And then try to connect/login (in case we haven't set up a
		// working connection previously in this request) and process
		// the remaining fields.
		$api = $this->_get_api();

		switch ($field)
		{
			case 'issue_type':
				if (isset($input['project'])) 
				{
					$data['default'] = arr::get($prefs, 'issue_type');
					$data['options'] = $this->_get_issue_types($api, $input['project']);
				}
				break;

			case 'project':
				$data['default'] = arr::get($prefs, 'project');
				$data['options'] = $this->_get_projects($api);
				break;

			case 'priority':
				$data['default'] = arr::get($prefs, 'priority');
				$data['options'] = $this->_get_priority($api);
				break;

			case 'category':
				if (isset($input['project']))
				{
					$data['default'] = arr::get($prefs, 'category');
					$data['options'] = $this->_get_categories($api,	$input['project']);
				}
				break;
		}
		return $data;
	}

	public function validate_push($context, $input)
	{
	}

	public function push($context, $input, array $path = [])
	{
		$api = $this->_get_api();
		return $api->add_issue($input, $path);
	}

	// *********************************************************
	// LOOKUP
	// *********************************************************

	public function lookup($defect_id)
	{
		$api = $this->_get_api();
		$issue = $api->get_issue($defect_id);

		$status_id = GI_DEFECTS_STATUS_OPEN;

		if (isset($issue->status))
		{
			$status = $issue->status->name;
		}
		else
		{
			$status = null;
		}

		if (isset($issue->description) && $issue->description)
		{
			$description = str::format(
				'<div class="monospace">{0}</div>',
				nl2br(
					html::link_urls(
						h($issue->description)
					)
				)
			);
		}
		else
		{
			$description = null;
		}

		// Add some important attributes for the issue such as the
		// current status and project.

		$attributes = array();

		if (isset($issue->issueType))
		{
			$attributes['IssueType'] = h($issue->issueType->name);
		}

		if ($status)
		{
			$attributes['Status'] = h($status);
		}

		if (isset($issue->issueKey))
		{
			$projectKey = $this->_get_project_from_id($api, $issue->projectId);

			// Add a link back to the project (issue list).
			$attributes['Project'] = str::format(
				'<a target="_blank" href="{0}projects/{1}">{2}</a>',
				a($this->_address),
				a($projectKey),
				h($projectKey)
			);
		}

		if (isset($issue->category))
		{
			$category_array = [];
			foreach ($issue->category as $category) {
				array_push($category_array, $category->name);
			}
			$attributes['Category'] = h(implode(', ', $category_array));
		}


		switch ($issue->status->id) {
			case 1:
			case 2:
				$status_id = GI_DEFECTS_STATUS_OPEN;
				break;
			case 3:
				$status_id = GI_DEFECTS_STATUS_RESOLVED;
				break;
			case 4:
				$status_id = GI_DEFECTS_STATUS_CLOSED;
				break;
			default:
				$status_id = GI_DEFECTS_STATUS_CLOSED;
		}

		return array(
			'id' => $defect_id,
			'url' => str::format(
				'{0}view/{1}',
				$this->_address,
				$defect_id
			),
			'title' => $issue->summary,
			'status_id' => $status_id,
			'status' => $status,
			'description' => $description,
			'attributes' => $attributes
		);
	}
}

/**
 * backlog API
 *
 * Wrapper class for the backlog API with functions for retrieving
 * projects, bugs etc. from a backlog installation.
 */
class Backlog_api
{
	private $_address;
	private $_api_key;
	private $_curl;

	/**
	 * Construct
	 *
	 * Initializes a new backlog API object. Expects the web address
	 * of the backlog installation including http or https prefix.
	 */
	public function __construct($address, $api_key)
	{
		$this->_address = str::slash($address);
		$this->_api_key = $api_key;
	}

	private function _throw_error($format, $params = null)
	{
		$args = func_get_args();
		$format = array_shift($args);

		if (count($args) > 0)
		{
			$message = str::formatv($format, $args);
		}
		else
		{
			$message = $format;
		}

		throw new BacklogException($message);
	}

	private function _send_command($method, $command, $content_type, $data = array(), $offset = null)
	{
		$url = $this->_address.$command."?apiKey=".$this->_api_key;
		return $this->_send_request($method, $url, $content_type, $data);
	}

	private function _send_request($method, $url, $content_type, $data)
	{
		if (!$this->_curl)
		{
			// Initialize the cURL handle. We re-use this handle to
			// make use of Keep-Alive, if possible.
			$this->_curl = http::open();
		}

		$response = http::request_ex(
			$this->_curl,
			$method,
			$url,
			array(
				'data' => $data,
				'api_key' => $this->_api_key,
				'headers' => array(
					'Content-Type' => $content_type
				)
			)
		);

		// In case debug logging is enabled, we append the data
		// we've sent and the entire request/response to the log.
		if (logger::is_on(GI_LOG_LEVEL_DEBUG))
		{
			logger::debugr('$data', $data);
			logger::debugr('$response', $response);
		}

		$obj = json::decode($response->content);

		if ($response->code != 200)
		{
			if ($response->code != 201) // Created
			{
				$this->_throw_error(
					'Invalid HTTP code ({0}). Please check your user/' .
					'password and that the API is enabled in backlog.',
					$response->code
				);
			}
		}
		return $obj;
	}

	/**
	 * Get Issue
	 */
	public function get_issue($issue_id)
	{
		$response = $this->_send_command('GET',	'api/v2/issues/'.urlencode($issue_id), 'application/json;charset=utf-8');
		return $response;
	}

	/**
	 * Get issueType
	 */
	public function get_issue_types($project_id)
	{
		$response = $this->_send_command('GET', 'api/v2/projects/'.$project_id.'/issueTypes', 'application/json;charset=utf-8');
		return $response;
	}

	/**
	 * Get Projects
	 */
	public function get_projects()
	{
		$response = $this->_send_command('GET','api/v2/projects', 'application/json;charset=utf-8');
		return $response;
	}

	/**
	 * Get ProjectKey from ProjectId
	 */
	public function get_project_from_id($project_id)
	{
		$response = $this->_send_command('GET', 'api/v2/projects', 'application/json;charset=utf-8');
		foreach ($response as $item) {
			if($item->id == $project_id){
				return $item->projectKey;
			}
		}
		return;
	}

	/**
	 * Get Categories
	 */
	public function get_categories($project_id)
	{
		$response = $this->_send_command('GET', 'api/v2/projects/'.$project_id.'/categories', 'application/json;charset=utf-8');
		return $response;
	}

	/**
	 * Get Priority
	 */
	public function get_priority()
	{
		$response = $this->_send_command('GET', 'api/v2/priorities', 'application/json;charset=utf-8');
		return $response;
	}

	/**
	 * Add Issue
	 */
	public function add_issue($options, array $paths)
	{
		$upload_params = "";
		foreach ($paths ?? [] as $path) {
			$upload = $this->_add_attachment($path);
			if (isset($upload->id)) {
				$upload_params .= "&attachmentId[]=" . (int)($upload->id);
			}
		}
		$data = 'projectId='. (int)$options['project'].
			'&summary='. urlencode($options['summary']).
			'&description='.urlencode($options['description']).
			'&issueTypeId=' . (int)$options['issue_type'] .
			'&priorityId=' . (int)$options['priority']. $upload_params;

		if(isset($options['category']))
		{
			$categoryIds = "";
			foreach ($options['category'] as $category) {
				$categoryIds .= "&categoryId[]=".(int)$category;
			}
			$data = $data.$categoryIds;
		}
		$response = $this->_send_command('POST', 'api/v2/issues', 'application/x-www-form-urlencoded', $data);
		return $response->issueKey;
	}

	/**
	 * Add Attachment
	 * Store attachment and return token and attacment id.
	 * 
	 * @param string $path User uploaded file path. 
	 * 
	 * @return object
	 */
	private function _add_attachment(string $path)
	{
		$path = str_replace("\\", "/", $path);
		$filename = basename($path);
		$url = $this->_address . 'api/v2/space/attachment?apiKey='. $this->_api_key;
		$file = new CURLFile($path, null, $filename);
		$postfields = array("file" => $file);
		$header = [
			'Content-Disposition: form-data; name="file"; filename=' . $filename.
			'Content-Type: application/octet-stream'
		];
		$ch = curl_init();
		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'POST');
		curl_setopt($ch, CURLOPT_HTTPHEADER, $header);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $postfields);
		curl_setopt($ch, CURLOPT_HEADER, true);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
		$response = curl_exec($ch);
		$header_size = curl_getinfo($ch, CURLINFO_HEADER_SIZE);
		$header = substr($response, 0, $header_size);
		$body = substr($response, $header_size);
		curl_close($ch);

		if (!preg_match('@HTTP/[0-9\.]+\s+([0-9]+)\s+(.*)\r\n@', $header, $matches)) {
			$this->_throw_error(
				'HTTP response something wrong. Status line not found. {0}',
				$header
			);
		}
		$status_code = (int)$matches[1];
		if($status_code != 200){
			$this->_throw_error(
				'Attachment is faild. {0}',
				$header
			);
		}
		$result = json_decode($body);
		return $result;
	}

	/*
	private function _log($data)
	{
		//$logfile = fopen("C:\\testrail\\custom\\defects\\backlog.log", "a");
		$logfile = fopen("/opt/testrail/logs/backlog.log", "a");
		$data_serialize = serialize($data);
		fwrite($logfile, "[LOG] ". $data_serialize . "\n");
		fclose($logfile);
	}
	*/
	
}

class BacklogException extends Exception
{
}
